"""Decorator functions."""

from collections.abc import Callable
from enum import Enum
from typing import Any, Protocol, TypeVar
from typing_extensions import deprecated

__all__ = ["enum_name_str", "implements"]


_T = TypeVar("_T")
_F = TypeVar("_F", bound=Callable[..., Any])


class IdentityFunction(Protocol[_T]):
    def __call__(self, __x: _T) -> _T: ...


@deprecated("Use 'typing_extensions.override' instead.")
class implements:  # pylint: disable=invalid-name
    """Mark a function as implementing an interface.

    :param interface: the interface that is implemented

    .. warning::
        This decorator is deprecated in favour of :func:`typing_extensions.override` instead.
    """

    def __init__(self, interface: type):
        self.interface = interface

    def __call__(self, func: _F) -> _F:
        """Take a function and return it unchanged."""
        super_method = getattr(self.interface, func.__name__, None)
        assert super_method is not None, f"'{func.__name__}' does not exist in {self.interface}"
        return func


E = TypeVar("E", bound=Enum)


@deprecated("Use `StrEnum` instead.")
def enum_name_str(enum_class: type[E]) -> type[E]:
    """Patch the __str__ method of an enum so that it returns the name."""
    # use the original __str__ method as __repr__
    enum_class.__repr__ = enum_class.__str__

    def __str__(self: Enum) -> str:
        return self.name.lower()

    enum_class.__str__ = __str__
    return enum_class



from collections.abc import Mapping, Sequence
from dataclasses import Field
from typing import (
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeGuard,
    TypeVar,
    get_type_hints,
    is_typeddict,
    runtime_checkable,
)

__all__ = ["Addable", "DataclassInstance", "Sized", "SizedDataset", "Subset", "is_td_instance"]

T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class Sized(Protocol[T_co]):
    def __len__(self) -> int: ...


@runtime_checkable
class Addable(Protocol[T_contra, T_co]):
    def __add__(self, other: T_contra, /) -> T_co: ...


@runtime_checkable
class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


TD = TypeVar("TD", bound=Mapping)


def is_td_instance(dict_: dict[str, Any], cls_: type[TD], *, strict: bool = False) -> TypeGuard[TD]:
    """``isinstance`` check for typed dictionaries.

    Returns ``True` if a dictionary conforms to a target ``TypedDict`` class.
    When ``strict` is ``False`` 'conforms' means that ``dict_`` contains all
    keys defined in ``cls_`` with values of the correct type -- additional keys
    are permitted. When ``True``, ``dict_`` must contain those keys **only**
    defined by ``cls_``.

    :param dict_: Dictionary to be checked.
    :param cls_: ``TypedDict`` class to compare ``dict_` against.

    :param strict: Whether to invalidate a ``dict_`` for containing additional
        keys not defined in ``cls_``

    :returns: ``True`` if ``dict_`` conforms to the target ``TypedDict`` class, ``cls_``,
        and ``False`` otherwise.
    """
    assert is_typeddict(cls_), "cls_ must be a TypedDict class."
    hints = get_type_hints(cls_)
    if strict and (len(dict_) != len(hints)):
        return False
    for key, type_ in hints.items():
        if (key not in dict_) or (not isinstance(dict_[key], type_)):
            return False
    return True


T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class SizedDataset(Protocol[T_co]):
    def __getitem__(self, index: int) -> T_co: ...

    def __len__(self) -> int: ...


D = TypeVar("D", bound=SizedDataset)


class Subset(Generic[D]):
    r"""
    Subset of a dataset at specified indices.

    :param dataset: The whole Dataset.
    :param indices: Indices in the whole set selected for subset.
    """

    dataset: D
    indices: Sequence[int]

    def __init__(self, dataset: D, indices: Sequence[int]):
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, index: int) -> Any:
        return self.dataset[self.indices[index]]

    def __len__(self) -> int:
        return len(self.indices)



from __future__ import annotations
from enum import Enum, auto
from typing import Generic, NamedTuple, TypeVar, overload

import torch
from torch import Tensor
import torch.distributions as td
import torch.nn.functional as F

from ranzen.misc import some, str_to_enum
from ranzen.torch.transforms.utils import sample_paired_indices

__all__ = [
    "MixUpMode",
    "RandomMixUp",
]


class MixUpMode(Enum):
    """An enum for the mix-up mode."""

    linear = auto()
    """linear mix-up"""
    geometric = auto()
    """geometric mix-up"""


class InputsTargetsPair(NamedTuple):
    inputs: Tensor
    targets: Tensor


LS = TypeVar("LS", td.Beta, td.Bernoulli, td.Uniform)


class RandomMixUp(Generic[LS]):
    r"""Apply mixup to tensors within a batch with some probability.

    PyTorch implemention of `mixup`_.
    This implementation allows for transformation of the the input in the absence
    of labels (this is relevant, for instance,to contrastive methods that use mixup to generate
    different views of samples to enable instance-discrimination) and additionally allows for
    different lambda-samplers, different methods for mixing up samples (linear vs. geometric)
    based on lambda, and selective pair-sampling. Furthermore, unlike the official implementation,
    samples are guaranteed not to be paired with themselves.

    .. note::
        This implementation randomly mixes images within a batch.

    :param lambda_sampler: The distribution from which to sample lambda (the mixup interpolation
        parameter).

    :param mode: Which mode to use to mix up samples: geometric or linear.

        .. note::
            The (weighted) geometric mean, enabled by ``mode=geometric``, is only valid for
            positive inputs.

    :param p: The probability with which the transform will be applied to a given sample.
    :param num_classes: The total number of classes in the dataset that needs to be specified if
        wanting to mix up targets that are label-enoded. Passing label-encoded targets without
        specifying ``num_classes`` will result in a RuntimeError.

    :param featurewise: Whether to sample sample feature-wise instead of sample-wise.

        .. note::
            If the ``lambda_sampler`` is a BernoulliDistribution, then featurewise sampling will
            always be enabled.

    :param inplace: Whether the transform should be performed in-place.
    :param generator: Pseudo-random-number generator to use for sampling. Note that
        :class:`torch.distributions.Distribution` does not accept such generator object and so
        the sampling procedure is only partially deterministic as a function of it.

    :raises ValueError: if ``p`` is not in the range [0, 1] or ``num_classes < 1``.

    .. _mixup:
        https://arxiv.org/abs/1904.00962v5
    """

    def __init__(
        self,
        lambda_sampler: LS,
        *,
        mode: MixUpMode | str = MixUpMode.linear,
        p: float = 1.0,
        num_classes: int | None = None,
        featurewise: bool = False,
        inplace: bool = False,
        generator: torch.Generator | None = None,
    ) -> None:
        super().__init__()
        self.lambda_sampler: LS = lambda_sampler
        if not 0 <= p <= 1:
            raise ValueError("'p' must be in the range [0, 1].")
        self.p = p
        if isinstance(mode, str):
            mode = str_to_enum(str_=mode, enum=MixUpMode)
        self.mode = mode
        if some(num_classes) and num_classes < 1:
            raise ValueError(f"{ num_classes } must be greater than 1.")
        self.num_classes = num_classes
        self.featurewise = featurewise or isinstance(lambda_sampler, td.Bernoulli)
        self.inplace = inplace
        self.generator = generator

    @classmethod
    def with_beta_dist(
        cls: type[RandomMixUp],
        alpha: float = 0.2,
        *,
        beta: float | None = None,
        mode: MixUpMode | str = MixUpMode.linear,
        p: float = 1.0,
        num_classes: int | None = None,
        inplace: bool = False,
        featurewise: bool = False,
        generator: torch.Generator | None = None,
    ) -> RandomMixUp[td.Beta]:
        """
        Instantiate a :class:`RandomMixUp` with a Beta-distribution sampler.

        :param alpha: 1st concentration parameter of the distribution. Must be positive
        :param beta:  2nd concentration parameter of the distribution. Must be positive.
            If ``None``, then the parameter will be set to ``alpha``.

        :param mode: Which mode to use to mix up samples: geometric or linear.

            .. note::
                The (weighted) geometric mean, enabled by ``mode=geometric``, is only valid for
                positive inputs.

        :param p: The probability with which the transform will be applied to a given sample.
        :param num_classes: The total number of classes in the dataset that needs to be specified if
            wanting to mix up targets that are label-enoded. Passing label-encoded targets without
            specifying ``num_classes`` will result in a RuntimeError.
        :param inplace: Whether the transform should be performed in-place.
        :param featurewise: Whether to sample sample feature-wise instead of sample-wise.
        :param generator: Pseudo-random-number generator to use for sampling. Note that
            :class:`torch.distributions.Distribution` does not accept such generator object and so
            the sampling procedure is only partially deterministic as a function of it.

        :return: A :class:`RandomMixUp` instance with ``lambda_sampler`` set to a  Beta-distribution
            with ``concentration1=alpha`` and ``concentration0=beta``.
        """
        beta = alpha if beta is None else beta
        lambda_sampler = td.Beta(concentration0=alpha, concentration1=beta)
        return cls(
            lambda_sampler=lambda_sampler,
            mode=mode,
            p=p,
            num_classes=num_classes,
            inplace=inplace,
            featurewise=featurewise,
            generator=generator,
        )

    @classmethod
    def with_uniform_dist(
        cls: type[RandomMixUp],
        low: float = 0.0,
        *,
        high: float = 1.0,
        mode: MixUpMode | str = MixUpMode.linear,
        p: float = 1.0,
        num_classes: int | None = None,
        inplace: bool = False,
        featurewise: bool = False,
        generator: torch.Generator | None,
    ) -> RandomMixUp[td.Uniform]:
        """
        Instantiate a :class:`RandomMixUp` with a uniform-distribution sampler.

        :param low: Lower range (inclusive).
        :param high: Upper range (inclusive).
        :param mode: Which mode to use to mix up samples: geometric or linear.

            .. note::
                The (weighted) geometric mean, enabled by ``mode=geometric``, is only valid for
                positive inputs.

        :param p: The probability with which the transform will be applied to a given sample.
        :param num_classes: The total number of classes in the dataset that needs to be specified if
            wanting to mix up targets that are label-enoded. Passing label-encoded targets without
            specifying ``num_classes`` will result in a RuntimeError unless ``num_classes`` is specified
            at call-time.
        :param inplace: Whether the transform should be performed in-place.
        :param featurewise: Whether to sample sample feature-wise instead of sample-wise.
        :param generator: Pseudo-random-number generator to use for sampling. Note that
            :class:`torch.distributions.Distribution` does not accept such generator object and so
            the sampling procedure is only partially deterministic as a function of it.

        :return: A :class:`RandomMixUp` instance with ``lambda_sampler`` set to a
            Uniform-distribution with ``low=low`` and ``high=high``.
        """
        lambda_sampler = td.Uniform(low=low, high=high)
        return cls(
            lambda_sampler=lambda_sampler,
            mode=mode,
            p=p,
            num_classes=num_classes,
            inplace=inplace,
            featurewise=featurewise,
            generator=generator,
        )

    @classmethod
    def with_bernoulli_dist(
        cls: type[RandomMixUp],
        prob_1: float = 0.5,
        *,
        mode: MixUpMode | str = MixUpMode.linear,
        p: float = 1.0,
        num_classes: int | None = None,
        inplace: bool = False,
        generator: torch.Generator | None,
    ) -> RandomMixUp[td.Bernoulli]:
        """
        Instantiate a :class:`RandomMixUp` with a Bernoulli-distribution sampler.

        :param prob_1: The probability of sampling 1.
        :param mode: Which mode to use to mix up samples: geometric or linear.

        .. note::
            The (weighted) geometric mean, enabled by ``mode=geometric``, is only valid for positive
            inputs.

        :param p: The probability with which the transform will be applied to a given sample.
        :param num_classes: The total number of classes in the dataset that needs to be specified if
            wanting to mix up targets that are label-enoded. Passing label-encoded targets without
            specifying ``num_classes`` will result in a RuntimeError.
        :param inplace: Whether the transform should be performed in-place.
        :param generator: Pseudo-random-number generator to use for sampling. Note that
            :class:`torch.distributions.Distribution` does not accept such generator object and so
            the sampling procedure is only partially deterministic as a function of it.

        :return: A :class:`RandomMixUp` instance with ``lambda_sampler`` set to a
            Bernoulli-distribution with ``probs=prob_1``.
        """
        lambda_sampler = td.Bernoulli(probs=prob_1)
        return cls(
            lambda_sampler=lambda_sampler,
            mode=mode,
            p=p,
            num_classes=num_classes,
            inplace=inplace,
            generator=generator,
        )

    def _mix(self, tensor_a: Tensor, *, tensor_b: Tensor, lambda_: Tensor) -> Tensor:
        lambda_c = 1 - lambda_
        if self.mode is MixUpMode.linear:
            return lambda_ * tensor_a + lambda_c * tensor_b
        return tensor_a**lambda_ * tensor_b**lambda_c

    @overload
    def _transform(
        self,
        inputs: Tensor,
        *,
        targets: Tensor,
        groups_or_edges: Tensor | None = ...,
        cross_group: bool = ...,
        num_classes: int | None = ...,
    ) -> InputsTargetsPair: ...

    @overload
    def _transform(
        self,
        inputs: Tensor,
        *,
        targets: None = ...,
        groups_or_edges: Tensor | None = ...,
        cross_group: bool = ...,
        num_classes: int | None = ...,
    ) -> Tensor: ...

    def _transform(
        self,
        inputs: Tensor,
        *,
        targets: Tensor | None = None,
        groups_or_edges: Tensor | None = None,
        cross_group: bool = False,
        num_classes: int | None = None,
    ) -> Tensor | InputsTargetsPair:
        sampled_index_pairs = sample_paired_indices(
            inputs=inputs,
            p=self.p,
            groups_or_edges=groups_or_edges,
            cross_group=cross_group,
            generator=self.generator,
        )
        if sampled_index_pairs is None:
            return inputs if targets is None else InputsTargetsPair(inputs=inputs, targets=targets)

        num_selected = len(sampled_index_pairs)
        anchor_indices = sampled_index_pairs.anchors
        match_indices = sampled_index_pairs.matches
        # Sample the mixing weights
        if self.featurewise:
            sample_shape = (num_selected, *inputs.shape[1:])
        else:
            sample_shape = (num_selected, *((1,) * (inputs.ndim - 1)))
        lambdas = self.lambda_sampler.sample(sample_shape=torch.Size(sample_shape)).to(
            inputs.device
        )

        if not self.inplace:
            inputs = inputs.clone()
        # Apply mixup to the inputs
        inputs[anchor_indices] = self._mix(
            tensor_a=inputs[anchor_indices], tensor_b=inputs[match_indices], lambda_=lambdas
        )
        if targets is None:
            return inputs

        # Targets are label-encoded and need to be one-hot encoded prior to mixup.
        if torch.atleast_1d(targets.squeeze()).ndim == 1:
            if num_classes is None:
                if self.num_classes is None:
                    raise RuntimeError(
                        f"{self.__class__.__name__} can only be applied to label-encoded targets "
                        "if 'num_classes' is specified."
                    )
                num_classes = self.num_classes
            elif num_classes < 1:
                raise ValueError(f"{ num_classes } must be a positive integer.")
            if num_classes > 2:
                targets = F.one_hot(targets.long(), num_classes=num_classes)
        elif not self.inplace:
            targets = targets.clone()
        # Targets need to be floats to be mixed up
        targets = targets.float()
        # Use the empirical mean of the lambdas for interpolating the targets if the lambdas
        # were sampled feasture-wise, else just use the lambdas as is.
        target_lambdas = lambdas = (
            lambdas.flatten(start_dim=1).mean(1) if self.featurewise else lambdas
        )
        # Add singular dimensions to lambdas for broadcasting
        target_lambdas = lambdas.view(num_selected, *((1,) * (targets.ndim - 1)))
        # Apply mixup to the targets
        targets[anchor_indices] = self._mix(
            tensor_a=targets[anchor_indices],
            tensor_b=targets[match_indices],
            lambda_=target_lambdas,
        )
        return InputsTargetsPair(inputs, targets)

    @overload
    def __call__(
        self,
        inputs: Tensor,
        *,
        targets: Tensor,
        groups_or_edges: Tensor | None = ...,
        cross_group: bool = ...,
        num_classes: int | None = ...,
    ) -> InputsTargetsPair: ...

    @overload
    def __call__(
        self,
        inputs: Tensor,
        *,
        targets: None = ...,
        groups_or_edges: Tensor | None = ...,
        cross_group: bool = ...,
        num_classes: int | None = ...,
    ) -> Tensor: ...

    def __call__(
        self,
        inputs: Tensor,
        *,
        targets: Tensor | None = None,
        groups_or_edges: Tensor | None = None,
        cross_group: bool = True,
        num_classes: int | None = None,
    ) -> Tensor | InputsTargetsPair:
        """
        :param inputs: The samples to apply mixup to.
        :param targets: The corresponding targets to apply mixup to. If the targets are
            label-encoded then the 'num_classes' attribute cannot be None.
        :param groups_or_edges: Labels indicating which group each sample belongs to or
            a boolean connectivity matrix encoding the permissibility of each possible pairing.
            In the case of the former, mixup pairs will be sampled in a cross-group fashion (only
            samples belonging to different groups will be paired for mixup) if ``cross_group``  is
            ``True`` and sampled in a within-group fashion (only sampled belonging to the same
            groups will be paired for mixup) otherwise.
        :param cross_group: Whether to sample mixup pairs in a cross-group (``True``) or
            within-group (``False``) fashion (see ``groups_or_edges``).
        :param num_classes: The total number of classes in the dataset that needs to be specified if
            wanting to mix up targets that are label-enoded. Passing label-encoded targets without
            specifying ``num_classes`` will result in a RuntimeError.

        :return: If target is None, the Tensor of mixup-transformed inputs. If target is not None, a
            namedtuple containing the Tensor of mixup-transformed inputs (inputs) and the
            corresponding Tensor of mixup-transformed targets (targets).
        """
        return self._transform(
            inputs=inputs,
            targets=targets,
            groups_or_edges=groups_or_edges,
            cross_group=cross_group,
            num_classes=num_classes,
        )



from collections.abc import Iterable, Iterator
from datetime import datetime
import random
from typing import Any, Optional, TypeVar, Union, overload
from typing_extensions import Self

import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor
import torch.nn as nn
from torch.types import Number

from ranzen.misc import some

__all__ = [
    "Event",
    "batchwise_pdist",
    "count_parameters",
    "inf_generator",
    "random_seed",
    "to_item",
    "to_numpy",
    "torch_eps",
]


def count_parameters(model: nn.Module) -> int:
    """Count all parameters (that have a gradient) in the given model"""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def random_seed(seed_value: int, *, use_cuda: bool) -> None:
    np.random.seed(seed_value)  # cpu vars
    torch.manual_seed(seed_value)  # cpu  vars
    random.seed(seed_value)  # Python
    if use_cuda:
        torch.cuda.manual_seed(seed_value)
        torch.cuda.manual_seed_all(seed_value)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


T = TypeVar("T")


def inf_generator(iterable: Iterable[T]) -> Iterator[T]:
    """Get DataLoaders in a single infinite loop.

    for i, (x, y) in enumerate(inf_generator(train_loader))

    :param iterable: An iterable that will be looped infinitely.
    :yield: Elements from the given iterable.
    :raises RuntimeError: If the given iterable is empty.
    """
    iterator = iter(iterable)
    # try to take one element to ensure that the iterator is not empty
    first_value = next(iterator, None)
    if first_value is not None:
        yield first_value
    else:
        raise RuntimeError("The given iterable is empty.")
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)


class Event:
    """Emulates torch.cuda.Event, but supports running on a CPU too.

    :example:
        >>> from ranzen.torch import Event
        >>> with Event() as event:
        >>>     y = some_nn_module(x)
        >>> print(event.time)
    """

    def __init__(self) -> None:
        self.time = 0.0
        self._cuda = torch.cuda.is_available()
        self._event_start: torch.cuda.Event | datetime

    def __enter__(self) -> Self:
        """Mark a time.

        Mimics torch.cuda.Event.
        """
        if self._cuda:
            self._event_start = torch.cuda.Event(enable_timing=True)  # pyright: ignore
            self._event_start.record()  # pyright: ignore
        else:
            self._event_start = datetime.now()
        return self

    def __exit__(self, *args: Any) -> None:
        if self._cuda:
            event_end = torch.cuda.Event(enable_timing=True)
            event_end.record(stream=torch.cuda.current_stream())
            torch.cuda.synchronize()
            assert isinstance(self._event_start, torch.cuda.Event)
            self.time = self._event_start.elapsed_time(event_end)
        else:
            assert isinstance(self._event_start, datetime)
            self.time = datetime.now().microsecond - self._event_start.microsecond

    def __repr__(self) -> str:
        return f"Event of duration: {self.time}"


def batchwise_pdist(x: Tensor, chunk_size: int = 1000, p_norm: float = 2.0) -> Tensor:
    """Compute pair-wise distance in batches.

    This is sometimes necessary because if you compute pdist directly, it doesn't fit into memory.

    :param x: Tensor of shape (N, F) where F is the number of features.
    :param chunk_size: Size of the chunks that are used to compute the pair-wise distance. Larger
        chunk size is probably faster, but may not fit into memory.
    :param p_norm: Which norm to use for the distance. Euclidean norm by default.
    :returns: All pair-wise distances in a tensor of shape (N, N).
    """
    chunks = torch.split(x, chunk_size)

    columns: list[Tensor] = []
    for chunk in chunks:
        shards = [torch.cdist(chunk, other_chunk, p_norm) for other_chunk in chunks]
        column = torch.cat(shards, dim=1)
        # the result has to be moved to the CPU; otherwise we'll run out of GPU memory
        columns.append(column.cpu())

        # free up memory
        for shard in shards:
            del shard
        del column
        torch.cuda.empty_cache()

    dist = torch.cat(columns, dim=0)

    # free up memory
    for column in columns:
        del column
    torch.cuda.empty_cache()
    return dist


DT = TypeVar("DT", bound=Union[np.number, np.bool_])


@overload
def to_numpy(tensor: Tensor, *, dtype: DT) -> npt.NDArray[DT]: ...


@overload
def to_numpy(tensor: Tensor, *, dtype: None = ...) -> npt.NDArray: ...


def to_numpy(
    tensor: Tensor, *, dtype: Optional[DT] | None = None
) -> Union[npt.NDArray[DT], npt.NDArray]:
    """
    Safely casts a tensor to a numpy ndarray of dtype ``dtype``
    if ``dtype`` is specified.

    :param tensor: Tensor to be cast to a ndarray.
    :param dtype: dtype of the cast-to ndarray. If ``None`` then
        the dtype will be inherited from ``tensor``.
    :returns: ``tensor`` cast to an ndarray of dtype ``dtype``
        if ``dtype`` is specified.
    """
    arr = tensor.detach().cpu().numpy()
    if some(dtype):
        arr.astype(dtype)
    return arr


def to_item(tensor: Tensor, /) -> Number:
    """
    Safely extracts the (int, float or bool) value from a single-element
    (scalar) tensor.

    :param tensor: Tensor to extract the item from.
    :returns: The value of the scalar ``tensor`` as a built-in float, int, bool.
    """
    return tensor.detach().cpu().item()


def torch_eps(tensor_or_dtype: Tensor | torch.dtype, /) -> float:
    """
    Retrieves the epsilon (the smallest representable number such that 1.0 + eps != 1.0.)
    value for the given dtype or the dtype of the given tensor.

    :param tensor_or_dtype: Tensor or :class:`torch.dtype` instance to retrieve
        the epislon value for.

    :returns: Epsilon value for the given dtype if ``tensor_or_dtype`` is an
        instance of :class:`torch.dtype` else the epsilon value of the dtype of the
        given tensor."""
    dtype = tensor_or_dtype.dtype if isinstance(tensor_or_dtype, Tensor) else tensor_or_dtype
    return torch.finfo(dtype).eps



from __future__ import annotations
from collections.abc import Iterable, Mapping, Sequence
import copy
from enum import Enum, auto
import functools
import operator
import sys
from typing import Any, Literal, TypeGuard, TypeVar, overload
from typing_extensions import Self

import numpy as np

from ranzen.types import Addable, SizedDataset, Subset

__all__ = [
    "AddDict",
    "Split",
    "Stage",
    "StrEnum",
    "default_if_none",
    "flatten_dict",
    "gcopy",
    "prop_random_split",
    "reduce_add",
    "some",
    "str_to_enum",
    "unwrap_or",
]


def flatten_dict(d: Mapping[str, Any], *, parent_key: str = "", sep: str = ".") -> dict[str, Any]:
    """Flatten a nested dictionary by separating the keys with `sep`.

    :param d: Dictionary to be flattened.
    :param parent_key: Key-prefix (separated from the key with 'sep') to use for top-level
        keys of the flattened dictionary.
    :param sep: Character to separate the parent keys from the child keys with at each level with.

    :returns: Flattened dictionary with keys capturing the nesting path as ``parent_key.child_key``,
        where 'parent_key' is defined recursively, with base value 'parent_key' as specified in the
        function call.
    """
    items: list[tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, Mapping):
            items.extend(flatten_dict(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


T = TypeVar("T")


@overload
def gcopy(obj: T, *, deep: bool = True, num_copies: None = ..., **kwargs: Any) -> T: ...


@overload
def gcopy(obj: T, *, deep: bool = True, num_copies: int, **kwargs: Any) -> list[T]: ...


def gcopy(
    obj: T, *, deep: bool = True, num_copies: int | None = None, **kwargs: Any
) -> T | list[T]:
    """Generalised (g) copy function.

    Allows for switching between deep and shallow copying within a single function
    as well as for the creation of multiple copies and for copying while simultaneously
    attribute-setting.

    :param obj: Object to be copied.
    :param deep: Whether to create deep (True) or shallow (False) copies.
    :param num_copies: Number of copies to create with 'None' being equivalent to 1.
    :param \\**kwargs: Key-word arguments specifying a name of an attribute and the
        new value to set it to in the copies.

    :returns: A copy or list of copies (if num_copies > 1) of the object 'obj'.

    :raises AttributeError: If an attribute specified in ``kwargs`` doesn't exist.
    """
    if num_copies is not None:
        return [gcopy(obj=obj, deep=deep, num_copies=None, **kwargs) for _ in range(num_copies)]
    obj_cp = copy.deepcopy(obj) if deep else copy.copy(obj)
    for attr, value in kwargs.items():
        if not hasattr(obj_cp, attr):
            raise AttributeError(
                f"Object of type '{type(obj_cp).__name__}' has no attribute '{attr}'."
            )
        setattr(obj_cp, attr, value)
    return obj_cp


E = TypeVar("E", bound=Enum)


def str_to_enum(str_: str | E, *, enum: type[E]) -> E:
    """Convert a string to an enum based on name instead of value.
    If the string is not a valid name of a member of the target enum,
    an error will be raised.

    :param str_: String to be converted to an enum member of type ``enum``.
    :param enum: Enum class to convert ``str_`` to.

    :returns: The enum member of type ``enum`` with name ``str_``.

    :raises TypeError: if the given string is not a valid name of a member of the target enum
    """
    if isinstance(str_, enum):
        return str_
    try:
        return enum[str_]  # type: ignore
    except KeyError:
        valid_ls = [mem.name for mem in enum]
        raise TypeError(
            f"'{str_}' is not a valid option for enum '{enum.__name__}'; must be one of {valid_ls}."
        )


if sys.version_info >= (3, 11):
    # will be available in python 3.11
    from enum import StrEnum
else:
    #
    # the following is copied straight from https://github.com/python/cpython/blob/3.11/Lib/enum.py
    #
    # DO NOT CHANGE THIS CODE!
    #
    class ReprEnum(Enum):
        """
        Only changes the repr(), leaving str() and format() to the mixed-in type.
        """

    _S = TypeVar("_S", bound="StrEnum")

    class StrEnum(str, ReprEnum):
        """
        Enum where members are also (and must be) strings
        """

        _value_: str

        def __new__(cls: type[_S], *values: str) -> _S:
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):  # pyright: ignore
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(v1 := values[1], str):  # pyright: ignore
                    raise TypeError("encoding must be a string, not %r" % (v1,))
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(v2 := values[2], str):  # pyright: ignore
                    raise TypeError("errors must be a string, not %r" % (v2))
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        def __str__(self) -> str:
            return str.__str__(self)

        def _generate_next_value_(  # type: ignore
            name: str,
            start: int,
            count: int,
            last_values: list[Any],
        ) -> str:
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()


_KT = TypeVar("_KT")
_VT = TypeVar("_VT", bound=Addable)
_VT2 = TypeVar("_VT2", bound=Addable)


class AddDict(dict[_KT, _VT], Addable[int | dict, dict]):
    """
    Extension of the built-in dictionary class that supports the use of the ``__add__`` operator for
    key-wise addition.

    :example:
        .. code-block:: python

            # Simple case of addition of integers.
            d1 = AddDict({"foo": 1, "bar": 2})
            d2 = {"foo": 3, "bar": 4}
            d1 + d2 # {'foo': 4, 'bar': 6}

            # Concatenation of lists
            d3 = AddDict({"foo": [1], "bar": [2]})
            d4 = {"foo": [3, 4], "bar": 4}
            d3 + d4 # {'foo': [1, 3, 4], 'bar': [2, 4]}
    """

    @overload
    def __add__(self, other: int) -> Self: ...

    @overload
    def __add__(self, other: dict[_KT, _VT2]) -> AddDict[_KT, _VT | _VT2]: ...

    def __add__(self, other: int | dict[_KT, _VT2]) -> Self | AddDict[_KT, _VT | _VT2]:
        # Allow ``other`` to be an integer, but specifying the identity function, for compatibility
        # with th 'no-default' version of``sum``.
        if isinstance(other, int):
            return self
        copy: AddDict[_KT, _VT | _VT2] = AddDict()
        copy.update(gcopy(self, deep=False))

        for key_o, value_o in other.items():
            if not isinstance(value_o, Addable):
                raise TypeError(f"Value of type '{type(value_o)}' is not addable.")
            if key_o in self:
                value_s = self[key_o]
                if not isinstance(value_s, Addable):
                    raise TypeError(f"Value of type '{type(value_s)}' is not addable.")
                try:
                    # Values are mutually addable (but not necessarily of the same type).
                    copy[key_o] = value_s + value_o
                except TypeError as e:
                    msg = (
                        f"Values of type '{type(value_s)}' and '{type(value_o)}' for key "
                        f"'{key_o}' are not mutuablly addable."
                    )
                    raise TypeError(msg) from e
            else:
                copy[key_o] = value_o
        return copy

    @overload
    def __radd__(self, other: int) -> Self: ...

    @overload
    def __radd__(self, other: dict[_KT, _VT2]) -> AddDict[_KT, _VT | _VT2]: ...

    def __radd__(self, other: int | dict[_KT, _VT2]) -> Self | AddDict[_KT, _VT | _VT2]:
        # Calling `__add__` directly because with the `+` syntax, pyright complains for some reason.
        return self.__add__(other)


A = TypeVar("A", bound=Addable)


def reduce_add(sequence: Iterable[A]) -> A:
    """
    Sum an iterable using functools.reduce and operator.add rather than the built-in sum operator
    to bypass need to specify an initial value or have the elements ``__add__`` operator be
    compatible with integers (0 being the default initial value).

    :param sequence: An iterable of addable instances, all of the same type (invariant).
    :returns: The sum of all elements in ``__iterable``.
    """
    return functools.reduce(operator.add, sequence)


def some(value: T | None, /) -> TypeGuard[T]:
    """
    Returns ``True`` if the input is **not** ``None``
    (that is, if the ``Optional`` monad contains some value).

    :param value: Value to be checked.
    :returns: ``True`` if ``value`` is **not** ``None`` else ``False``.
    """
    return value is not None


def unwrap_or(value: T | None, /, *, default: T) -> T:
    """
    Returns the input if the input is **not** None else the specified
    ``default`` value.

    :param value: Input to be unwrapped and returned if not ``None``.
    :param default: Default value to use if ``value`` is ``None``.
    :returns: ``default`` if ``value`` is ``None`` otherwise ``value``.
    """
    return default if value is None else value


default_if_none = unwrap_or


class Stage(StrEnum):
    FIT = auto()
    VALIDATE = auto()
    TEST = auto()


class Split(StrEnum):
    TRAIN = auto()
    VAL = auto()
    TEST = auto()


D = TypeVar("D", bound=SizedDataset)


@overload
def prop_random_split(
    dataset_or_size: D,
    *,
    props: Sequence[float] | float,
    as_indices: Literal[False] = ...,
    seed: int | None = ...,
    reproducible: bool = ...,
) -> list[Subset[D]]: ...


@overload
def prop_random_split(
    dataset_or_size: SizedDataset,
    *,
    props: Sequence[float] | float,
    as_indices: Literal[True],
    seed: int | None = ...,
    reproducible: bool = ...,
) -> list[list[int]]: ...


@overload
def prop_random_split(
    dataset_or_size: int,
    *,
    props: Sequence[float] | float,
    as_indices: bool = ...,
    seed: int | None = ...,
    reproducible: bool = ...,
) -> list[list[int]]: ...


@overload
def prop_random_split(
    dataset_or_size: D | int,
    *,
    props: Sequence[float] | float,
    as_indices: bool = ...,
    seed: int | None = ...,
    reproducible: bool = ...,
) -> list[Subset[D]] | list[list[int]]: ...


def prop_random_split(
    dataset_or_size: D | int,
    *,
    props: Sequence[float] | float,
    as_indices: bool = False,
    seed: int | None = None,
    reproducible: bool = False,
) -> list[Subset[D]] | list[list[int]]:
    """Splits a dataset based on proportions rather than on absolute sizes

    :param dataset_or_size: Dataset or size (length) of the dataset to split.
    :param props: The fractional size of each subset into which to randomly split the data.
        Elements must be non-negative and sum to 1 or less; if less then the size of the final
        split will be computed by complement.

    :param as_indices: If ``True`` the raw indices are returned instead of subsets constructed
        from them when `dataset_or_len` is a dataset. This means that when `dataset_or_len`
        corresponds to the length of a dataset, this argument has no effect and
        the function always returns the split indices.

    :param seed: The PRNG used for determining the random splits.

    :param reproducible: If ``True``, use a generator which is reproducible across machines,
        operating systems, and Python versions.

    :returns: Random subsets of the data of the requested proportions.

    :raises ValueError: If the dataset does not have a ``__len__`` method or sum(props) > 1.
    """
    if isinstance(dataset_or_size, int):
        len_ = dataset_or_size
    else:
        if not hasattr(dataset_or_size, "__len__"):
            raise ValueError(
                "Split proportions can only be computed for datasets with __len__ defined."
            )
        len_ = len(dataset_or_size)

    if isinstance(props, (float, int)):
        props = [props]
    sum_ = np.sum(props)
    if (sum_ > 1.0) or any(prop < 0 for prop in props):
        raise ValueError("Values for 'props` must be positive and sum to 1 or less.")
    section_sizes = [round(prop * len_) for prop in props]
    if (current_len := sum(section_sizes)) < len_:
        section_sizes.append(len_ - current_len)

    if reproducible:
        if seed is None:
            raise ValueError("Must specify seed for reproducible split.")
        # MT19937 isn't the best random number generator, but it's reproducible, so we're using it.
        generator = np.random.Generator(np.random.MT19937(seed))
    else:
        generator = np.random.default_rng(seed)
    indices = np.arange(sum(section_sizes))
    generator.shuffle(indices)  # Shuffle the indices in-place.

    splits = [
        indices[offset - length : offset].tolist()
        for offset, length in zip(np.cumsum(section_sizes), section_sizes)
    ]

    if as_indices or isinstance(dataset_or_size, int):
        return splits
    return [Subset(dataset_or_size, indices=split) for split in splits]



import operator
from typing import NamedTuple
import warnings

import torch
from torch import Tensor

from ranzen.torch.sampling import batched_randint

__all__ = [
    "sample_paired_indices",
    "PairedIndices",
]


class PairedIndices(NamedTuple):
    anchors: Tensor
    matches: Tensor

    def __len__(self) -> int:
        return len(self.anchors)


def sample_paired_indices(
    inputs: Tensor,
    *,
    p: float,
    groups_or_edges: Tensor | None = None,
    cross_group: bool = False,
    generator: torch.Generator | None = None,
) -> PairedIndices | None:
    batch_size = len(inputs)
    # If the batch is singular or the sampling probability is 0 there's nothing to do.
    if (batch_size == 1) or (p == 0):
        return None
    if p < 1:
        # Sample a mask determining which samples in the batch are to be transformed
        selected = torch.rand(batch_size, device=inputs.device, generator=generator) < p
        num_selected = int(selected.count_nonzero())
        indices = selected.nonzero(as_tuple=False).long().flatten()
    # if p >= 1 then the transform is always applied and we can skip
    # the above step
    else:
        num_selected = batch_size
        indices = torch.arange(batch_size, device=inputs.device, dtype=torch.long)

    if groups_or_edges is None:
        # Sample the mixup pairs with the guarantee that a given sample will
        # not be paired with itself
        offset = torch.randint(
            low=1,
            high=batch_size,
            size=(num_selected,),
            device=inputs.device,
            dtype=torch.long,
            generator=generator,
        )
        pair_indices = (indices + offset) % batch_size
    else:
        groups_or_edges = groups_or_edges.squeeze()
        if groups_or_edges.ndim == 1:
            if len(groups_or_edges) != batch_size:
                raise ValueError(
                    "The number of elements in 'groups_or_edges' should match the size of "
                    "dimension 0 of 'inputs'."
                )

            groups = groups_or_edges.view(batch_size, 1)  # [batch_size]
            # Compute the pairwise indicator matrix, indicating whether any two samples
            # belong to the same group (0) or different groups (1)
            comp = operator.ne if cross_group else operator.eq
            connections: Tensor = comp(groups[indices], groups.t())  # [num_selected, batch_size]
            if not cross_group:
                # Fill the diagonal with 'False' to prevent self-matches.
                connections.fill_diagonal_(fill_value=False)
            # For each sample, compute how many other samples there are that belong
            # to a different group.
        elif groups_or_edges.ndim == 2:
            if groups_or_edges.dtype is not torch.bool:
                raise ValueError(
                    "If 'groups_or_edges' is a matrix, it must have dtype 'torch.bool'."
                )
            connections = groups_or_edges[indices]
            # Fill the diagonal with 'False' to prevent self-matches.
            connections.fill_diagonal_(fill_value=False)
        else:
            raise ValueError(
                "'groups_or_edges' must be a vector denoting group membership or a boolean-type"
                "groups_or_edges matrix with elements denoting the permissibility of each"
                "possible pairing in 'inputs'."
            )
        degrees = connections.count_nonzero(dim=1)  # [num_selected]
        is_connected = degrees.nonzero().squeeze(-1)
        num_selected = len(is_connected)
        if num_selected < len(degrees):
            warnings.warn(
                (
                    "One or more samples without valid pairs according to "
                    "the connectivity defined by 'groups_or_edges'."
                ),
                RuntimeWarning,
            )
        # If there are no valid pairs (all vertices are isolated), there's nothing to do.
        if num_selected == 0:
            return None
        # Update the tensors to account for pairless samples.
        indices = indices[is_connected]
        degrees = degrees[is_connected]
        connections = connections[is_connected]
        # Sample the mixup pairs in accordance with the connectivity matrix.
        # This can be efficiently done as follows:
        # 1) Sample uniformly from {0, ..., diff_group_count - 1} to obtain the groupwise pair indices.
        # This involves first drawing samples from the standard uniform distribution, rescaling them to
        # [-1/(2*diff_group_count), diff_group_count + (1/(2*diff_group_count)], and then clamping them
        # to [0, 1], making it so that 0 and diff_group_count have the same probability of being drawn
        # as any other value. The uniform samples are then mapped to indices by multiplying by
        # diff_group_counts and rounding. 'randint' is unsuitable here because the groups aren't
        # guaranteed to have equal cardinality (using it to sample from the cyclic group,
        # Z / diff_group_count Z, as above, leads to biased sampling).
        rel_pair_indices = batched_randint(degrees, generator=generator)
        # 2) Convert the row-wise indices into row-major indices, considering only
        # only the postive entries in the rows.
        rel_pair_indices[1:] += degrees.cumsum(dim=0)[:-1]
        # 3) Finally, map from relative indices to absolute ones.
        _, abs_pos_inds = connections.nonzero(as_tuple=True)
        pair_indices = abs_pos_inds[rel_pair_indices]
    return PairedIndices(
        anchors=indices,
        matches=pair_indices,
    )



from typing import Any
from typing_extensions import override

import torch
from torch import Tensor
from torch.optim.optimizer import Optimizer

from .common import LossClosure

__all__ = ["SAM"]


class SAM(Optimizer):
    """
    Implements the 'Sharpness Aware Minimization' (SAM) algorithm introducued
    in `Sharpness Aware Minimization`_ along with the adaptive variant of it
    introduced in `ASAM`_.

    SAM seeks parameters that lie in neighborhoods having uniformly low loss (rather than
    parameters that only themselves have low loss value). The adaptive variant of the
    algorithm addresses the original algorithm's sensitivity to parameter re-scaling
    that can lead to weakening of the connection between sharpness and generalization gap.

    :param base_optimizer: Base optimizer for SAM.
    :param rho: Neighborhood size.
    :param adaptive: Whether to use the adaptive variant of the algorithm.

    :raises ValueError: if ``rho`` is negative.

    :example:
        .. code-block:: python

            # Use AdamW as the base optimizer.
            base_optimizer = AdamW(model.parameters())
            # Wrap the base optimizer in SAM.
            optimizer = SAM(base_optimizer)

            # Closure required for recomputing the loss after computing epsilon(w).
            def _closure():
              return loss_function(logits=model(input), targets=targets)

            loss = _closure()
            loss.backward()

            optimizer.step(closure=_closure)
            optimizer.zero_grad()

    .. _Sharpness Aware Minimization:
        https://arxiv.org/abs/2010.01412
    .. _ASAM:
        https://arxiv.org/abs/2102.11600
    """

    def __init__(
        self,
        base_optimizer: Optimizer,
        *,
        rho: float = 0.05,
        adaptive: bool = True,
    ):
        if rho < 0.0:
            raise ValueError(f"Invalid rho value: {rho}. (Should be non-negative)")

        defaults = dict(rho=rho, adaptive=adaptive)
        self.base_optimizer = base_optimizer
        self.param_groups = self.base_optimizer.param_groups
        super(SAM, self).__init__(params=base_optimizer.param_groups, defaults=defaults)

    @torch.no_grad()  # pyright: ignore
    def _first_step(self, *, zero_grad: bool = False) -> None:
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None:
                    continue
                self.state[p]["old_p"] = p.data.clone()
                e_w = (p.pow(2) if group["adaptive"] else 1.0) * p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()  # pyright: ignore
    def _second_step(self, *, zero_grad: bool = False) -> None:
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                p.data = self.state[p]["old_p"]  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()  # pyright: ignore
    @override
    def step(self, closure: LossClosure) -> Tensor:  # type: ignore
        r"""Performs a single optimization step.

        :param closure: A closure that reevaluates the model and returns the loss.
        :returns: loss returned by the closure if ``closure`` is not ``None`` else ``None``.
        """
        self._first_step(zero_grad=True)
        with torch.enable_grad():
            loss = closure()
            loss.backward()
        self._second_step()
        return loss

    def _grad_norm(self) -> Tensor:
        shared_device = self.param_groups[0]["params"][
            0
        ].device  # put everything on the same device, in case of model parallelism
        norm = torch.norm(
            torch.stack(
                [
                    ((torch.abs(p) if group["adaptive"] else 1.0) * p.grad)
                    .norm(p=2)
                    .to(shared_device)
                    for group in self.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p="fro",
        )
        return norm

    @override
    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads the optimizer state.

        :param state_dict: optimizer state. Should be an object returned
            from a call to :meth:`state_dict`.
        """
        super().load_state_dict(state_dict)
        self.base_optimizer.param_groups = self.param_groups



from abc import abstractmethod
from collections import defaultdict
from collections.abc import Iterator, Sequence, Sized
from dataclasses import dataclass
from enum import Enum, auto
import math
from typing import Final, Generic, NewType, Optional, TypeVar, Union, cast
from typing_extensions import Self, override

import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor
from torch.utils.data import Sampler

from ranzen.misc import prop_random_split as prop_random_split  # noqa: PLC0414
from ranzen.misc import some, str_to_enum
from ranzen.types import SizedDataset as SizedDataset  # noqa: PLC0414
from ranzen.types import Subset as Subset  # noqa: PLC0414

__all__ = [
    "ApproxStratBatchSampler",
    "BatchSamplerBase",
    "GreedyCoreSetSampler",
    "SequentialBatchSampler",
    "SizedDataset",
    "StratifiedBatchSampler",
    "Subset",
    "TrainTestSplit",
    "TrainingMode",
    "WeightedBatchSampler",
    "prop_random_split",
    "stratified_split_indices",
]


_S = TypeVar("_S")


@dataclass(frozen=True)
class TrainTestSplit(Generic[_S]):
    train: _S
    test: _S

    def __iter__(self) -> Iterator[_S]:
        yield from (self.train, self.test)


def stratified_split_indices(
    labels: Tensor | npt.NDArray[np.int_] | Sequence[int],
    *,
    default_train_prop: float,
    train_props: dict[int, float] | None = None,
    seed: int | None = None,
) -> TrainTestSplit[list[int]]:
    """Splits the data into train/test sets conditional on super- and sub-class labels.

    :param labels: Tensor, array or sequence encoding the label associated with each sample.
    :param default_train_prop: Proportion of samples for a given to sample for
        the training set for those y-s combinations not specified in ``train_props``.

    :param train_props: Proportion of each group  to sample for the training set.
        If ``None`` then the function reduces to a simple random split of the data.

    :param seed: PRNG seed to use for sampling.

    :returns: Train-test split.

    :raises ValueError: If a value in ``train_props`` is not in the range [0, 1] or if a key is not
        present in ``group_ids``.
    """
    if not isinstance(labels, Tensor):
        labels = torch.as_tensor(labels, dtype=torch.long)
    # Initialise the random-number generator
    generator = torch.default_generator if seed is None else torch.Generator().manual_seed(seed)
    groups, label_counts = labels.unique(return_counts=True)
    train_props_all = dict.fromkeys(groups.tolist(), default_train_prop)

    if train_props is not None:
        for label, train_prop in train_props.items():
            if not 0 <= train_prop <= 1:
                raise ValueError(
                    "All splitting proportions specified in 'train_props' must be in the "
                    "range [0, 1]."
                )
            if label not in groups:
                raise ValueError(f"No samples belonging to the group in 'group_ids'.")
            train_props_all[label] = train_prop

    # Shuffle the samples before sampling
    perm_inds = torch.randperm(len(labels), generator=generator)
    labels_perm = labels[perm_inds]

    sort_inds = labels_perm.sort(dim=0, stable=True).indices
    thresholds = cast(
        Tensor, (torch.as_tensor(tuple(train_props_all.values())) * label_counts).round().long()
    )
    thresholds = torch.stack([thresholds, label_counts], dim=-1)
    thresholds[1:] += label_counts.cumsum(0)[:-1].unsqueeze(-1)

    train_test_inds = sort_inds.tensor_split(thresholds.flatten()[:-1], dim=0)
    train_inds = perm_inds[torch.cat(train_test_inds[0::2])].tolist()
    test_inds = perm_inds[torch.cat(train_test_inds[1::2])].tolist()

    return TrainTestSplit(train=train_inds, test=test_inds)


class TrainingMode(Enum):
    """An enum for the training mode."""

    epoch = auto()
    """epoch-based training"""
    step = auto()
    """step-based training"""


class BatchSamplerBase(Sampler[list[int]]):
    def __init__(self, epoch_length: int | None = None) -> None:
        self.epoch_length: Final[int | None] = epoch_length

    @override
    @abstractmethod
    def __iter__(self) -> Iterator[list[int]]: ...

    def __len__(self) -> float | int:
        """The number of samples drawn.
        If epoch_length is None then the sampler has no length defined and will
        be sampled from infinitely. However, in such cases, __len__ still needs to
        be defined for downstream compatibility (e.g. with PyTorch Lightning) and for
        this it suffices to simply return 'inf'.
        """
        if self.epoch_length is None:
            return math.inf
        return self.epoch_length


def _check_generator(generator: torch.Generator | None) -> torch.Generator:
    """If the generator is None, randomly initialise a generator object."""
    if generator is None:
        generator = torch.Generator()
        generator = generator.manual_seed(int(torch.empty((), dtype=torch.int64).random_().item()))
    return generator


def num_batches_per_epoch(num_samples: int, *, batch_size: int, drop_last: bool = False) -> int:
    rounding_fn = math.floor if drop_last else math.ceil
    return rounding_fn(num_samples / batch_size)


class SequentialBatchSampler(BatchSamplerBase):
    r"""Infinitely samples elements sequentially, always in the same order.

    This is useful for enabling iteration-based training.
    Note that unlike torch's SequentialSampler which is an ordinary sampler that yields independent
    sample indexes, this is a BatchSampler, requiring slightly different treatment when used with a
    DataLoader.

    :param data_source: Object of the same size as the data to be sampled from.
    :param batch_size: How many samples per batch to load.
    :param training_mode: The training mode to use (epoch vs. step).
    :param shuffle: Set to ``True`` to have the data reshuffled at every epoch.
    :param drop_last: Set to ``True`` to drop the last incomplete batch,
    :param generator: Pseudo-random-number generator to use for shuffling the dataset.

    :example:
        >>> batch_sampler = InfSequentialBatchSampler(data_source=train_data, batch_size=100, shuffle=True)
        >>> train_loader = DataLoader(train_data, batch_sampler=batch_sampler, shuffle=False, drop_last=False) # drop_last and shuffle need to be False
        >>> train_loader_iter = iter(train_loader)
        >>> for _ in range(train_iters):
        >>>     batch = next(train_loader_iter)
    """

    def __init__(
        self,
        data_source: Sized,
        *,
        batch_size: int,
        training_mode: TrainingMode | str = TrainingMode.step,
        shuffle: bool = True,
        drop_last: bool = False,
        generator: torch.Generator | None = None,
    ) -> None:
        self.data_source = data_source
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.generator = generator
        if isinstance(training_mode, str):
            training_mode = str_to_enum(str_=training_mode, enum=TrainingMode)
        self.training_mode = training_mode
        if self.training_mode is TrainingMode.epoch:
            epoch_length = num_batches_per_epoch(
                num_samples=len(self.data_source),
                batch_size=self.batch_size,
                drop_last=self.drop_last,
            )
        else:
            epoch_length = None
        super().__init__(epoch_length=epoch_length)

    def _generate_idx_seq(self, generator: torch.Generator) -> Tensor:
        """Generate a random sequence of unique indexes."""
        if self.shuffle:
            return torch.randperm(len(self.data_source), generator=generator)
        return torch.arange(len(self.data_source))

    def _batch_indexes(self, indexes: Tensor) -> Sequence[Tensor]:
        """Split the indexes into batches."""
        return indexes.split(self.batch_size)

    @override
    def __iter__(self) -> Iterator[list[int]]:
        generator = _check_generator(self.generator)
        batched_idxs_iter = iter(self._batch_indexes(self._generate_idx_seq(generator=generator)))
        # Iterate until some stopping criterion is reached
        while True:
            batch_idxs = next(batched_idxs_iter, None)
            if (batch_idxs is None) or (len(batch_idxs) < self.batch_size):
                if self.epoch_length is None:
                    new_idx_seq = self._generate_idx_seq(generator=generator)
                    if (batch_idxs is not None) and (not self.drop_last):
                        # Rather than dropping the last batch if it is incomplete or simply using it,
                        # incomplete as it may be, we take the alternative approach of concatenating the surplus
                        # batch to the beginning of the next generation of indexes
                        new_idx_seq = torch.cat([batch_idxs, new_idx_seq])
                    batched_idxs_iter = iter(self._batch_indexes(new_idx_seq))
                else:
                    if (batch_idxs is not None) and (not self.drop_last):
                        yield batch_idxs.tolist()
                    break
            else:
                yield batch_idxs.tolist()


class BaseSampler(Enum):
    """An enum for the base-sampler to use for StratifiedBatchSampler."""

    random = auto()
    """random sampler"""
    sequential = auto()
    """sequential sampler"""


class StratifiedBatchSampler(BatchSamplerBase):
    r"""Samples equal proportion of elements from ``[0,..,len(group_ids)-1]``.

    To drop certain groups, set their multiplier to 0.

    :param group_ids: A sequence of group IDs, not necessarily contiguous.

    :param num_samples_per_group: Number of samples to draw per group. Note that if a multiplier is > 1
        then effectively more samples will be drawn for that group.

    :param multipliers: An optional dictionary that maps group IDs to multipliers. If a multiplier
        is greater than 1, the corresponding group will be sampled at twice the rate as the other
        groups. If a multiplier is 0, the group will be skipped.

    :param base_sampler: The base sampling strategy to use (sequential vs. random).

    :param training_mode: The training mode to use (epoch vs. step).

    :param replacement: if ``True``, samples are drawn with replacement. If not, they are drawn
        without replacement, which means that when a sample index is drawn for a row, it cannot be
        drawn again for that row.

    :param shuffle: Whether to shuffle the subsets of the data after each pass (only applicable when
        the base_sampler is set to ``sequential``).

    :param drop_last: Set to ``True`` to drop the last (on a per-group basis) incomplete batch.
    :param generator: Pseudo-random-number generator to use for shuffling the dataset.
    :raises ValueError: If ``num_samples_per_group`` is non-positive, if ``replacement`` is not a
        bool, or if there are not enough samples in a group to sample ``num_samples_per_group``.

    :example:
        >>> list(StratifiedSampler([0, 0, 0, 0, 1, 1, 2], 10, replacement=True))
        [3, 5, 6, 3, 5, 6, 0, 5, 6]
        >>> list(StratifiedSampler([0, 0, 0, 0, 1, 1, 2], 10, replacement=True, multiplier={2: 2}))
        [3, 4, 6, 6, 3, 5, 6, 6, 1, 5, 6, 6]
        >>> list(StratifiedSampler([0, 0, 0, 0, 1, 1, 1, 2, 2], 7, replacement=False))
        [2, 6, 7, 0, 5, 8]
    """

    def __init__(
        self,
        group_ids: Sequence[int],
        *,
        num_samples_per_group: int,
        multipliers: dict[int, int] | None = None,
        base_sampler: BaseSampler | str = BaseSampler.sequential,
        training_mode: TrainingMode | str = TrainingMode.step,
        replacement: bool = True,
        shuffle: bool = False,
        drop_last: bool = True,
        generator: torch.Generator | None = None,
    ) -> None:
        if (
            not isinstance(num_samples_per_group, int)
            or isinstance(num_samples_per_group, bool)
            or num_samples_per_group <= 0
        ):
            raise ValueError(
                f"num_samples_per_group should be a positive integer; got {num_samples_per_group}"
            )
        if not isinstance(replacement, bool):
            raise ValueError(
                f"replacement should be a boolean value, but got replacement={replacement}"
            )
        if isinstance(base_sampler, str):
            base_sampler = str_to_enum(str_=base_sampler, enum=BaseSampler)
        if isinstance(training_mode, str):
            training_mode = str_to_enum(str_=training_mode, enum=TrainingMode)

        self.num_samples_per_group = num_samples_per_group
        multipliers_ = {} if multipliers is None else multipliers

        group_ids_t = torch.as_tensor(group_ids, dtype=torch.int64)
        # find all unique IDs
        groups: list[int] = group_ids_t.unique().tolist()

        # get the indexes for each group separately and compute the effective number of groups
        groupwise_idxs: list[tuple[Tensor, int]] = []
        num_groups_effective = 0
        for group in groups:
            # Idxs needs to be 1 dimensional
            idxs = (group_ids_t == group).nonzero(as_tuple=False).view(-1)
            multiplier = multipliers_.get(group, 1)
            assert isinstance(multiplier, int) and multiplier >= 0, "multiplier has to be >= 0"
            groupwise_idxs.append((idxs, multiplier))
            num_groups_effective += multiplier

            if not replacement and len(idxs) < num_samples_per_group * multiplier:
                raise ValueError(
                    f"Not enough samples in group {group} to sample {num_samples_per_group}."
                )

        self.groupwise_idxs = groupwise_idxs
        self.num_groups_effective = num_groups_effective
        self.batch_size = self.num_groups_effective * self.num_samples_per_group
        self.sampler = base_sampler
        self.replacement = replacement
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.generator = generator
        self.training_mode = training_mode

        if self.training_mode is TrainingMode.epoch:
            # We define the length of the sampler to be the maximum number of steps
            # needed to do a complete pass of a group's data
            groupwise_epoch_length = [
                num_batches_per_epoch(
                    num_samples=len(idxs),
                    batch_size=mult * num_samples_per_group,
                    drop_last=self.drop_last,
                )
                for idxs, mult in self.groupwise_idxs
            ]
            # Sort the groupwise-idxs by their associated epoch-length
            sorted_idxs_desc = np.argsort(groupwise_epoch_length)[::-1]
            self.groupwise_idxs = [self.groupwise_idxs[idx] for idx in sorted_idxs_desc]
            max_epoch_length = groupwise_epoch_length[sorted_idxs_desc[0]]
        else:
            max_epoch_length = None

        super().__init__(epoch_length=max_epoch_length)

    def _sequential_sampler(self, generator: torch.Generator) -> Iterator[list[int]]:
        samplers_and_idxs = [
            (
                iter(
                    SequentialBatchSampler(
                        data_source=group_idxs,
                        batch_size=self.num_samples_per_group * multiplier,
                        shuffle=self.shuffle,
                        generator=generator,
                        training_mode=self.training_mode
                        if self.training_mode is TrainingMode.epoch and (group_num == 0)
                        else TrainingMode.step,  # group-idxs are sorted by epoch-length
                    )
                ),
                group_idxs,
            )
            for group_num, (group_idxs, multiplier) in enumerate(self.groupwise_idxs)
            # Skip any groups with a non-positive multiplier
            if (multiplier > 0)
        ]
        sampled_idxs: list[int]
        # 'step' mode is enabled, making the sampling procedure very simple
        # because there's no need to special case the last batch
        if self.epoch_length is None:
            while True:
                sampled_idxs = []
                for sampler, group_idxs in samplers_and_idxs:
                    sampled_idxs.extend(group_idxs[next(sampler)].tolist())
                yield sampled_idxs
        # 'epoch' mode is enabled - handling the last batch is quite involved
        # as we need to preserve the ratio between the groups prescribed by the
        # multipliers
        else:
            # Factor by which to reduce the batch by - only relevant for the
            # last batch and is computed as the ratio of the number of
            # drawn for the final batch to the group-specific batch size
            # for the group with the longest epoch-length
            batch_reduction_factor: float | None = None
            for step in range(1, self.epoch_length + 1):
                sampled_idxs = []
                for group_num, (sampler, group_idxs) in enumerate(samplers_and_idxs):
                    idxs_of_idxs = next(sampler)
                    # The groups are ordered by epoch-length so we only need to check the first group
                    # (being the one that dictates the length of a epoch for the whole sampler)
                    if group_num == 0:
                        if step == self.epoch_length:
                            batch_reduction_factor = len(idxs_of_idxs) / (
                                self.num_samples_per_group * self.groupwise_idxs[group_num][1]
                            )
                            # The batch is incomplete and drop-last is enabled - terminate the iteration
                            if self.drop_last and (not batch_reduction_factor):
                                return
                    elif batch_reduction_factor is not None:
                        # Subsample the indexes according to the batch-reduction-factor
                        reduced_sample_count = round(len(idxs_of_idxs) * batch_reduction_factor)
                        idxs_of_idxs = idxs_of_idxs[:reduced_sample_count]
                    # Collate the indexes
                    sampled_idxs.extend(group_idxs[idxs_of_idxs].tolist())

                yield sampled_idxs

    def _random_sampler(self, generator: torch.Generator) -> Iterator[list[int]]:
        while True:
            sampled_idxs: list[Tensor] = []
            for group_idxs, multiplier in self.groupwise_idxs:
                if self.replacement:
                    for _ in range(multiplier):
                        # sampling with replacement:
                        # just sample enough random numbers to fill the quota
                        idxs_of_idxs = torch.randint(
                            low=0,
                            high=len(group_idxs),
                            size=(self.num_samples_per_group,),
                            generator=generator,
                        )
                        sampled_idxs.append(group_idxs[idxs_of_idxs])
                else:
                    # sampling without replacement:
                    # first shuffle the indexes and then take as many as we need
                    shuffled_idx = group_idxs[torch.randperm(len(group_idxs), generator=generator)]
                    # all elements in `sampled_idx` have to have the same size,
                    # so we split the tensor in equal-sized parts and then take as many as we need
                    chunks = torch.split(shuffled_idx, self.num_samples_per_group)
                    sampled_idxs += list(chunks[:multiplier])
            yield torch.cat(sampled_idxs, dim=0).tolist()

    @override
    def __iter__(self) -> Iterator[list[int]]:
        generator = _check_generator(self.generator)
        if self.sampler is BaseSampler.random:
            return self._random_sampler(generator=generator)
        else:
            return self._sequential_sampler(generator=generator)


class GreedyCoreSetSampler(BatchSamplerBase):
    r"""Constructs batches from 'oversampled' batches through greedy core-set approximation.

    Said approximation takes the form of the furtherst-frist traversal (FFT) algorithm.

    :param embeddings: Embedded dataset from which to sample the core-sets according;
        the order of the embeddings, v, must match the order of the dataset
        (i.e. f(x_i) = v_i, for embedding function f and inputs x)

    :param batch_size: Budget for the core-set,

    :param oversampling_factor: How many times larger than the budget the batch to be sampled from
        should be.

    :param generator: Pseudo-random-number generator to use for shuffling the dataset.
    """

    def __init__(
        self,
        embeddings: Tensor,
        *,
        batch_size: int,
        oversampling_factor: int,
        generator: torch.Generator | None = None,
    ) -> None:
        self.oversampling_factor = oversampling_factor
        self.embeddings = embeddings.flatten(start_dim=1).detach().cpu()
        self.budget = batch_size
        self._num_oversampled_samples = min(
            self.budget * self.oversampling_factor, len(self.embeddings)
        )
        self.generator = generator

        super().__init__(epoch_length=None)

    def _get_dists(self, batch_idxs: Tensor) -> Tensor:
        batch = self.embeddings[batch_idxs]
        dist_mat = batch @ batch.t()
        sq = dist_mat.diagonal().view(batch.size(0), 1)
        return -2 * dist_mat + sq + sq.t()

    @override
    def __iter__(self) -> Iterator[list[int]]:
        generator = _check_generator(self.generator)
        # iterative forever (until some externally defined stopping-criterion is reached)
        while True:
            # First sample the 'oversampled' batch from which to construct the core-set
            os_batch_idxs = torch.randperm(len(self.embeddings), generator=generator)[
                : self._num_oversampled_samples
            ]
            # Compute the euclidean distance between all pairs in said batch
            dists = self._get_dists(os_batch_idxs)
            # Mask indicating whether a sample is still yet to be sampled (1=unsampled, 0=sampled)
            # - updating a mask is far more efficnet than reconstructing the list of unsampled
            # indexes every iteration (however, we do have to be careful about the 'meta-indexing'
            # it introduces)
            unsampled_m = torch.ones_like(os_batch_idxs, dtype=torch.bool)
            # there's no obvious decision rule for seeding the core-set so we just select the first
            # point arbitrarily, moving it from the unsampled pool to the sampled one by setting
            # its corresponding mask-value to 0
            sampled_idxs = [int(os_batch_idxs[0])]
            unsampled_m[0] = 0

            # Begin the furthest-first traversal algorithm
            while len(sampled_idxs) < self.budget:
                # p := argmax min_{i\inB}(d(x, x_i)); i.e. select the point which maximizes the
                # minimum squared Euclidean-distance to all previously selected points
                # NOTE: The argmax index is relative to the unsampled indexes
                rel_idx = torch.argmax(torch.min(dists[~unsampled_m][:, unsampled_m], dim=0).values)
                # Retiieve the index corresponding to the previously-computed argmax index
                to_sample = os_batch_idxs[unsampled_m][rel_idx]
                sampled_idxs.append(int(to_sample))
                # Update the mask, which corresponds to moving the sampled index from the unsampled
                # pool to the sampled pool
                unsampled_m[unsampled_m.nonzero()[rel_idx]] = 0

            yield sampled_idxs


class WeightedBatchSampler(BatchSamplerBase):
    r"""Implements a batch-sampler version of :class:`torch.utils.data.WeightedRandomSampler`.

    :param weights:  A sequence or tensor of weights, not necessarily summing to one.
    :param batch_size: Number of samples to draw per batch/iteration.
    :param replacement: If ``True``, samples are drawn with replacement.
        If not, they are drawn without replacement, which means that when a sample index is drawn
        for a row, it cannot be drawn again for that row.

    :param generator: Pseudo-random-number generator to use for randomly sampling indexes.

    :raises ValueError: If ``batch_size`` is non-positive or is greater than the number of weights
        when ``replacement=False``
    """

    def __init__(
        self,
        weights: Sequence[float] | Tensor,
        *,
        batch_size: int,
        replacement: bool = True,
        generator: torch.Generator | None = None,
    ) -> None:
        if batch_size < 1:
            raise ValueError(
                f"batch_size must be a positive integer, but got batch_size={batch_size}."
            )
        if (not replacement) and ((len_ := len(weights)) < batch_size):
            raise ValueError(
                "batch_size cannot be greater than the number of weights if sampling "
                f"without replacement, but got batch_size={batch_size} and weights of length {len_}"
            )
        self.weights = torch.as_tensor(weights, dtype=torch.float32)
        self.batch_size = batch_size
        self.replacement = replacement
        self.generator = generator

        super().__init__(epoch_length=None)

    @classmethod
    def from_labels(
        cls,
        labels: Sequence[int] | Tensor,
        *,
        batch_size: int,
        replacement: bool = True,
        generator: torch.Generator | None = None,
    ) -> Self:
        """
        Instantiate a :class:`~WeightedBatchSampler` from a sequenece or tensor of ints, where
        ``weights`` is computed using the inverse frequencies of the values in ``labels``.

        :param labels: Labels from which to compute the sample weights from; should be of length
            equal to the size of the associated dataset being indexed.

        :param batch_size: Number of samples to draw per batch/iteration.
        :param replacement: If ``True``, samples are drawn with replacement.
            If not, they are drawn without replacement, which means that when a
            sample index is drawn for a row, it cannot be drawn again for that row.

        :param generator: Pseudo-random-number generator to use for randomly sampling indexes.

        :returns: A :class:`~WeightedBatchSampler` instance with ``weights`` computed using the
            inverse frequencies of the values in ``labels``.

        :raises ValueError: If ``labels`` is a tensor and does not have dtype :class:`torch.long`.
        """
        if isinstance(labels, Tensor):
            if labels.dtype is not torch.long:
                raise ValueError("labels must have dtype 'long' ('int64').")
        else:
            labels = torch.as_tensor(labels, dtype=torch.long)
        _, inverse, counts = labels.unique(return_inverse=True, return_counts=True)
        counts_r = counts.reciprocal()
        sample_weights = counts_r / counts_r.sum()
        weights = sample_weights[inverse].squeeze()
        return cls(
            weights=weights,
            batch_size=batch_size,
            replacement=replacement,
            generator=generator,
        )

    @override
    def __iter__(self) -> Iterator[list[int]]:
        generator = _check_generator(self.generator)
        # Iterate until some stopping criterion is reached
        while True:
            yield torch.multinomial(
                self.weights,
                num_samples=self.batch_size,
                replacement=self.replacement,
                generator=generator,
            ).tolist()


S = NewType("S", int)  # type meant to help with documentation
Y = NewType("Y", int)


class ApproxStratBatchSampler(BatchSamplerBase):
    r"""Approximate Stratified Batch Sampler.

    Essentially, we’re doing: :math:`x\sim P(x|s,y)` where :math:`s\sim \text{uniform}(S|y)`. That
    is, we iterate over all classes y and uniformly sample a subgroup s, and then we sample a
    datapoint from that s-y combination.

    You have to either specify ``num_samples_per_group`` or ``num_samples_per_class`` (but not
    both).

    If ``num_samples_per_group`` is given, this faithfully implements the :math:`\pi` function. This means
    that for those classes which have "full s-support" (all subgroups are present), we don’t sample
    a subgroup but iterate over each subgroup one-by-one. We take ``num_samples_per_group`` samples
    from each s-y combination.

    On the other hand, if ``num_samples_per_class`` is given, then classes with full s-support are
    not given special treatment. We always sample as many subgroups as are specified in
    ``num_samples_per_class`` and then take a single datapoint from each of these s-y combinations.

    :param class_labels: List-like object with the class labels.
    :param subgroup_labels: List-like object with the subgroup labels.
    :param num_samples_per_group: How many samples to take per s-y group. Cannot be specified
        together with ``num_samples_per_class``.
    :param num_samples_per_class: How many samples to take per y class. Cannot be specified
        together with ``num_samples_per_group``.
    :param training_mode: Iteration-based vs epoch-based.
    :param generator: Torch generator for random numbers.
    :raises ValueError: If not exactly one of ``num_samples_per_group`` and
        ``num_samples_per_class`` is specified.
    """

    def __init__(
        self,
        class_labels: Sequence[int],
        subgroup_labels: Sequence[int],
        *,
        num_samples_per_group: Optional[int] = None,
        num_samples_per_class: Optional[int] = None,
        training_mode: TrainingMode = TrainingMode.step,
        generator: Union[torch.Generator, None] = None,
    ) -> None:
        if some(num_samples_per_group) and some(num_samples_per_class):
            raise ValueError(
                "Specify either `num_samples_per_group` or `num_samples_per_class` but not both"
            )
        elif (num_samples_per_group is None) and (num_samples_per_class is None):
            raise ValueError("Specify one of `num_samples_per_group` or `num_samples_per_class`")

        assert len(class_labels) == len(subgroup_labels), "labels should have the same length"

        class_labels_t = torch.as_tensor(class_labels, dtype=torch.int64)
        subgroup_labels_t = torch.as_tensor(subgroup_labels, dtype=torch.int64)
        # find all unique labels
        classes = class_labels_t.unique().tolist()
        subgroups = subgroup_labels_t.unique().tolist()
        # cast to nice-looking types
        classes = cast(list[Y], classes)
        subgroups = cast(list[S], subgroups)

        # get the indexes for each group separately and store them in a hierarchical dict
        groupwise_idxs: defaultdict[Y, list[Tensor]] = defaultdict(list)
        for class_ in classes:
            for subgroup in subgroups:
                # Idxs needs to be 1 dimensional
                idxs = (
                    ((subgroup_labels_t == subgroup) & (class_labels_t == class_))
                    .nonzero(as_tuple=False)
                    .view(-1)
                )
                num_samples = len(idxs)

                if num_samples == 0:
                    continue
                groupwise_idxs[class_].append(idxs)

                if some(num_samples_per_group) and (num_samples < num_samples_per_group):
                    raise ValueError(
                        f"Not enough samples in group (s={subgroup}, y={class_}) "
                        f"to sample {num_samples_per_group} (available: {num_samples})."
                    )

        self.groupwise_idxs = groupwise_idxs
        self.generator = generator
        if some(num_samples_per_group):
            # In each batch, we want the full list of subgroups
            self.num_subgroup_samples = len(subgroups)
            self.num_samples_per_group = num_samples_per_group
            self.classes_with_full_support: set[Y] = {
                y
                for y, subgroup_idxs in groupwise_idxs.items()
                if len(subgroup_idxs) == self.num_subgroup_samples
            }
        elif some(num_samples_per_class):
            self.num_subgroup_samples = num_samples_per_class
            self.num_samples_per_group = 1  # Take one sample for each sampled s
            self.classes_with_full_support = set()  # No special-casing for full-support classes
        else:
            raise RuntimeError("shouldn’t happen")

        self.batch_size = len(classes) * self.num_subgroup_samples * self.num_samples_per_group

        if training_mode is TrainingMode.epoch:
            # some groups have fewer samples than others
            # we want to know how many batches to sample to cover even the biggest group
            if some(num_samples_per_group):
                max_epoch_length = max(
                    num_batches_per_epoch(
                        num_samples=len(idxs), batch_size=num_samples_per_group, drop_last=False
                    )
                    for subgroup_idxs in self.groupwise_idxs.values()
                    for idxs in subgroup_idxs
                )
            elif some(num_samples_per_class):
                # (the following calculation is a bit sketchy, but I think it's fine)
                max_epoch_length = max(
                    num_batches_per_epoch(
                        num_samples=max(len(idxs) for idxs in subgroup_idxs) * len(subgroup_idxs),
                        batch_size=num_samples_per_class,
                        drop_last=False,
                    )
                    for subgroup_idxs in self.groupwise_idxs.values()
                )
            else:
                raise RuntimeError("shouldn’t happen")
        else:
            max_epoch_length = None

        super().__init__(epoch_length=max_epoch_length)

    @override
    def __iter__(self) -> Iterator[list[int]]:
        generator = self.generator
        if generator is None:
            generator = torch.Generator()
            generator = generator.manual_seed(
                int(torch.empty((), dtype=torch.int64).random_().item())
            )

        while True:
            sampled_idxs: list[Tensor] = []
            for y, subgroupwise_idxs in self.groupwise_idxs.items():
                if y in self.classes_with_full_support:
                    # just take samples from each subgroup
                    sampled_idxs.extend(
                        self._take_samples_per_group(idxs, generator) for idxs in subgroupwise_idxs
                    )
                else:
                    # first sample (with replacement) the required number of subgroups
                    subgroups = torch.randint(
                        low=0,
                        high=len(subgroupwise_idxs),
                        size=(self.num_subgroup_samples,),
                        generator=generator,
                    )
                    # then take samples from each
                    sampled_idxs.extend(
                        self._take_samples_per_group(subgroupwise_idxs[i], generator)
                        for i in subgroups.tolist()
                    )
            yield torch.cat(sampled_idxs, dim=0).tolist()

    def _take_samples_per_group(self, tensor: Tensor, generator: torch.Generator) -> Tensor:
        # first shuffle and then take as many as we need
        shuffled = tensor[torch.randperm(len(tensor), generator=generator)]
        return shuffled[: self.num_samples_per_group]



"""Functions for dealing with hydra."""

from collections.abc import Iterator, MutableMapping, Sequence
from contextlib import contextmanager
import dataclasses
from dataclasses import MISSING, asdict, is_dataclass
from enum import Enum
import shlex
from typing import Any, Final, cast, get_args, get_type_hints
from typing_extensions import deprecated

from hydra.core.config_store import ConfigStore
from hydra.core.hydra_config import HydraConfig
from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf

from ranzen.types import DataclassInstance

__all__ = [
    "GroupRegistration",
    "SchemaRegistration",
    "as_pretty_dict",
    "reconstruct_cmd",
    "recursively_instantiate",
    "prepare_for_logging",
    "register_hydra_config",
]

NEED: Final = "there should be"
IF: Final = "if an entry has"


def _clean_up_dict(obj: Any) -> Any:
    """Convert enums to strings and filter out _target_."""
    if isinstance(obj, MutableMapping):
        return {key: _clean_up_dict(value) for key, value in obj.items() if key != "_target_"}
    elif isinstance(obj, Enum):
        return str(f"{obj.name}")
    elif OmegaConf.is_config(obj):  # hydra stores lists as omegaconf.ListConfig, so we convert here
        return OmegaConf.to_container(obj, resolve=True, enum_to_str=True)
    return obj


def as_pretty_dict(data_class: DataclassInstance) -> dict:
    """Convert dataclass to a pretty dictionary."""
    return _clean_up_dict(asdict(data_class))


def reconstruct_cmd() -> str:
    """Reconstruct the python command that was used to start this program."""
    internal_config = HydraConfig.get()
    program = internal_config.job.name + ".py"
    args = internal_config.overrides.task
    return shlex.join([program] + OmegaConf.to_container(args))  # type: ignore[operator]


@deprecated("Use _recursive_=True instead.")
def recursively_instantiate(
    hydra_config: DictConfig, *, keys_to_exclude: Sequence[str] = ()
) -> dict[str, Any]:
    return {
        str(k): instantiate(v, _convert_="partial")
        for k, v in hydra_config.items()
        if k not in ("_target_",) + tuple(keys_to_exclude)
    }


class GroupRegistration:
    """Helper for registering a group in hydra."""

    def __init__(self, cs: ConfigStore, *, group_name: str, package: str):
        self._cs = cs
        self._group_name = group_name
        self._package = package

    def add_option(self, config_class: type, *, name: str) -> None:
        """Register a schema as an option for this group."""
        self._cs.store(group=self._group_name, name=name, node=config_class, package=self._package)


class SchemaRegistration:
    """Register hydra schemas.

    :example:
        >>> sr = SchemaRegistration()
        >>> sr.register(Config, path="experiment_schema")
        >>> sr.register(TrainerConf, path="trainer/trainer_schema")
        >>>
        >>> with sr.new_group("schema/data", target_path="data") as group:
        >>>    group.add_option(CelebaDataConf, name="celeba")
        >>>    group.add_option(WaterbirdsDataConf, name="waterbirds")
    """

    def __init__(self) -> None:
        self._cs = ConfigStore.instance()

    def register(self, config_class: type, *, path: str) -> None:
        """Register a schema."""
        if "." in path:
            raise ValueError(f"Separate path with '/' and not '.': {path}")

        parts = path.split("/")
        name = parts[-1]
        package = ".".join(parts[:-1])
        self._cs.store(name=name, node=config_class, package=package)

    @contextmanager
    def new_group(self, group_name: str, *, target_path: str) -> Iterator[GroupRegistration]:
        """Return a context manager for a new group."""
        package = target_path.replace("/", ".")
        yield GroupRegistration(self._cs, group_name=group_name, package=package)


def prepare_for_logging(hydra_config: DictConfig, *, enum_to_str: bool = True) -> dict[str, Any]:
    """Takes a hydra config dict and makes it prettier for logging.

    Things this function does: turn enums to strings, resolve any references, mark entries with
    their type.
    """
    raw_config = OmegaConf.to_container(
        hydra_config, throw_on_missing=True, enum_to_str=enum_to_str, resolve=True
    )
    assert isinstance(raw_config, dict)
    raw_config = cast(dict[str, Any], raw_config)
    return {
        f"{key}/{OmegaConf.get_type(dict_).__name__}"  # type: ignore
        if isinstance(dict_ := hydra_config[key], DictConfig)
        else key: value
        for key, value in raw_config.items()
    }


def register_hydra_config(
    main_cls: type, groups: dict[str, dict[str, type]], schema_name: str = "config_schema"
) -> None:
    """Check the given config and store everything in the ConfigStore.

    This function performs two tasks: 1) make the necessary calls to `ConfigStore`
    and 2) run some checks over the given config and if there are problems, try to give a nice
    error message.

    :param main_cls: The main config class; can be dataclass or attrs.
    :param groups: A dictionary that defines all the variants. The keys of top level of the
        dictionary should corresponds to the group names, and the keys in the nested dictionaries
        should correspond to the names of the options.
    :param schema_name: Name of the main schema. This name has to appear in the defaults list in the
        main config file.
    :raises ValueError: If the config is malformed in some way.
    :raises RuntimeError: If hydra itself is throwing an error.

    :example:
        .. code-block:: python

            @dataclass
            class DataModule:
                root: Path = Path()

            @dataclass
            class LinearModel:
                dim: int = 256

            @dataclass
            class CNNModel:
                kernel: int = 3

            @dataclass
            class Config:
                dm: DataModule = dataclasses.field(default_factory=DataModule)
                model: Any

            groups = {"model": {"linear": LinearModel, "cnn": CNNModel}}
            register_hydra_config(Config, groups)
    """
    assert isinstance(main_cls, type), "`main_cls` has to be a type."
    if not is_dataclass(main_cls):
        raise ValueError(f"The config class {main_cls.__name__} should be a dataclass.")
    entries = dataclasses.fields(main_cls)
    try:
        types = get_type_hints(main_cls)
    except NameError as exc:
        raise ValueError(
            f"Can't resolve type hints from the config class: `{main_cls.__name__}`."
        ) from exc

    for entry in entries:
        typ = types[entry.name]
        if typ == Any:
            if (group := groups.get(entry.name)) is not None:
                for var_name, var_class in group.items():
                    if not is_dataclass(var_class):
                        raise ValueError(
                            f"All variants should be dataclasses: type `{var_class.__name__}` "
                            f"of variant `{entry.name}={var_name}` is not a dataclass."
                        )
            else:
                raise ValueError(f"{IF} type `Any`, {NEED} variants: `{entry.name}`")
            if entry.default is not MISSING or entry.default_factory is not MISSING:
                raise ValueError(f"{IF} type `Any`, {NEED} no default value: `{entry.name}`")
        else:
            if is_dataclass(typ):
                if entry.default is MISSING and entry.default_factory is MISSING:
                    if (group := groups.get(entry.name)) is not None:
                        for var_name, var_class in group.items():
                            if not issubclass(var_class, typ):
                                raise ValueError(
                                    f"All variants should be subclasses of their entry's type: type"
                                    f" `{var_class.__name__}` of variant `{entry.name}={var_name}` "
                                    f"is not a subclass of `{typ.__name__}`."
                                )
                    else:
                        raise ValueError(
                            f"{IF} a dataclass type, "
                            f"{NEED} a default value or registered variants: `{entry.name}`. "
                            "You can specify a default value with `field(default_factory=...)`."
                        )
                else:
                    if entry.name in groups:
                        raise ValueError(
                            f"Can't have both a default value and variants: `{entry.name}`."
                        )
            elif entry.name in groups:
                raise ValueError(
                    f"Entry `{entry.name}` has registered variants, but its type, "
                    f"`{entry.type.__name__}`, is not a dataclass."
                )

    cs = ConfigStore.instance()
    cs.store(node=main_cls, name=schema_name)
    for group, entries in groups.items():
        for var_name, var_type in entries.items():
            if (bases := getattr(var_type, "__orig_bases__", None)) is not None:
                if len(bases) > 0 and len(get_args(bases[0])) > 0:
                    raise ValueError(
                        f"Can't register a dataclass with generic base class: `{var_type.__name__}`"
                        f" with base class `{bases[0].__name__}`."
                    )
            try:
                cs.store(node=var_type, name=var_name, group=group)
            except Exception as exc:
                raise RuntimeError(f"{main_cls=}, {var_type=}, {var_name=}, {group=}") from exc



# Most of the code here has been copied from:
# https://github.com/pytorch/fairseq/blob/main/fairseq/optim/adafactor.py
from collections.abc import Iterable, Sequence
import math
from typing import TypedDict, cast

import torch
from torch import Tensor
from torch.optim import Optimizer

from .common import LossClosure

__all__ = ["Adafactor"]


class ParamGroup(TypedDict):
    params: Iterable[Tensor]
    lr: float
    eps: tuple[float, float]
    clipping_threshold: float
    decay_rate: float
    beta1: float | None
    weight_decay: float
    multiply_by_parameter_scale: bool
    warmup_init: bool
    relative_step: bool


class ParamState(TypedDict):
    lr: float
    rms: Tensor
    step: int
    exp_avg: Tensor
    exp_avg_sq: Tensor
    exp_avg_sq_row: Tensor
    exp_avg_sq_col: Tensor


class Adafactor(Optimizer):
    """Implements Adafactor algorithm.

    This implementation is based on: `Adafactor: Adaptive Learning Rates with
    Sublinear Memory Cost <see https://arxiv.org/abs/1804.04235>`_. Note that
    this optimizer internally adjusts the learning rate depending on the
    ``multiply_by_parameter_scale``, ``relative_step`` and ``warmup_init``
    options. To use a manual (external) learning rate schedule you should set
    ``multiply_by_parameter_scale=False`` and ``relative_step=False``.

    :param params: iterable of parameters to optimize or dicts defining parameter groups

    :param lr: learning rate. If ``None``, a time-dependent learning rate will instead be computed.

    :param eps: regularization constants for square gradient
        and parameter scale respectively.

    :param clipping_threshold: threshold of root mean square of final gradient update.

    :param decay_rate: coefficient used to compute running averages of square gradient.

    :param beta1: coefficient used for computing running averages of gradient.

    :param weight_decay: weight decay coefficient.
    :param multiply_by_parameter_scale: if True, learning rate is scaled by
        root mean square of parameter.

    :param warmup_init: time-dependent learning rate computation depends on
        whether warm-up initialization is being used.
    """

    param_groups: list[ParamGroup]  # type: ignore

    def __init__(
        self,
        params: Iterable[Tensor],
        *,
        lr: float | None = None,
        eps: tuple[float, float] = (1e-30, 1e-3),
        clipping_threshold: float = 1.0,
        decay_rate: float = 0.8,
        beta1: float | None = None,
        weight_decay: float = 0.0,
        multiply_by_parameter_scale: bool = False,
        warmup_init: bool = False,
    ) -> None:
        relative_step = lr is None
        defaults = dict(
            lr=lr,
            eps=eps,
            clipping_threshold=clipping_threshold,
            decay_rate=decay_rate,
            beta1=beta1,
            weight_decay=weight_decay,
            multiply_by_parameter_scale=multiply_by_parameter_scale,
            relative_step=relative_step,
            warmup_init=warmup_init,
        )
        super().__init__(params, defaults)

    def _get_lr(self, param_group: ParamGroup, param_state: ParamState) -> float:
        rel_step_sz = param_group["lr"]
        if param_group["relative_step"]:
            min_step = 1e-6 * param_state["step"] if param_group["warmup_init"] else 1e-2
            rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state["step"]))
        param_scale = 1.0
        if param_group["multiply_by_parameter_scale"]:
            param_scale = max(param_group["eps"][1], param_state["rms"])  # type: ignore
        return param_scale * rel_step_sz

    def _get_options(
        self, param_group: ParamGroup, param_shape: Sequence[int]
    ) -> tuple[bool, bool]:
        factored = len(param_shape) >= 2
        use_first_moment = param_group["beta1"] is not None
        return factored, use_first_moment

    def _rms(self, tensor: Tensor) -> Tensor:
        return tensor.norm(p=2) / (tensor.numel() ** 0.5)

    def _approx_sq_grad(self, *, exp_avg_sq_row: Tensor, exp_avg_sq_col: Tensor) -> Tensor:
        r_factor = (
            (exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1, keepdim=True)).rsqrt_().unsqueeze(-1)
        )
        c_factor = exp_avg_sq_col.unsqueeze(-2).rsqrt()
        return torch.mul(r_factor, c_factor)

    def step(self, closure: LossClosure | None = None) -> Tensor | None:  # type: ignore
        """Performs a single optimization step.
        :param closure: A closure that reevaluates the model and returns the loss.
        :returns: loss returned by the closure if ``closure`` is not ``None`` else ``None``.
        :raises RuntimeError: if gradients are sparse.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.dtype in {torch.float16, torch.bfloat16}:
                    grad = grad.float()
                if grad.is_sparse:
                    raise RuntimeError("Adafactor does not support sparse gradients.")

                state = self.state[p]
                grad_shape = grad.shape

                factored, use_first_moment = self._get_options(group, grad_shape)
                # State Initialization
                if len(state) == 0:
                    state["step"] = 0

                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state["exp_avg"] = torch.zeros_like(grad)
                    if factored:
                        state["exp_avg_sq_row"] = torch.zeros(grad_shape[:-1]).to(grad)
                        state["exp_avg_sq_col"] = torch.zeros(grad_shape[:-2] + grad_shape[-1:]).to(
                            grad
                        )
                    else:
                        state["exp_avg_sq"] = torch.zeros_like(grad)

                    state["rms"] = torch.zeros(())
                else:
                    if use_first_moment:
                        state["exp_avg"] = state["exp_avg"].to(grad)
                    if factored:
                        state["exp_avg_sq_row"] = state["exp_avg_sq_row"].to(grad)
                        state["exp_avg_sq_col"] = state["exp_avg_sq_col"].to(grad)
                    else:
                        state["exp_avg_sq"] = state["exp_avg_sq"].to(grad)

                p_data_fp32 = p.data
                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p_data_fp32 = p_data_fp32.float()

                state["step"] += 1
                state["rms"] = self._rms(p_data_fp32)
                group["lr"] = self._get_lr(group, state)

                beta2t = 1.0 - math.pow(state["step"], -group["decay_rate"])
                update = (grad**2) + group["eps"][0]
                if factored:
                    exp_avg_sq_row = state["exp_avg_sq_row"]
                    exp_avg_sq_col = state["exp_avg_sq_col"]

                    exp_avg_sq_row.mul_(beta2t).add_(update.mean(dim=-1), alpha=1.0 - beta2t)
                    exp_avg_sq_col.mul_(beta2t).add_(update.mean(dim=-2), alpha=1.0 - beta2t)

                    # Approximation of exponential moving average of square of gradient
                    update = self._approx_sq_grad(
                        exp_avg_sq_row=exp_avg_sq_row, exp_avg_sq_col=exp_avg_sq_col
                    )
                    update.mul_(grad)
                else:
                    exp_avg_sq = state["exp_avg_sq"]

                    exp_avg_sq.mul_(beta2t).add_(update, alpha=1.0 - beta2t)
                    update = exp_avg_sq.rsqrt().mul_(grad)

                update.div_((self._rms(update) / group["clipping_threshold"]).clamp_(min=1.0))
                update.mul_(group["lr"])

                if use_first_moment:
                    momentum = cast(float, group["beta1"])
                    exp_avg = state["exp_avg"]
                    exp_avg.mul_(momentum).add_(update, alpha=1 - momentum)
                    update = exp_avg

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(p_data_fp32, alpha=-group["weight_decay"] * group["lr"])

                p_data_fp32.add_(-update)

                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p.data.copy_(p_data_fp32)

        return loss
